// ---------------------------------------------------------------------
// Time-stamp: <Tuesday, 10.12.2002, 13:29:06; edited by bauke>
// 
// Tina's random number generators TRNG
//
// lagged Fibonacci generator
//
// Copyright (C) 2001, 2002 Heiko Bauke
//
// heiko.bauke@physik.uni-magdeburg.de
//
// TRNG is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation. This program
// is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
//
// ---------------------------------------------------------------------

#ifndef FIBONACCI_H
#define FIBONACCI_H

#include <trng.h>

// This is an example for extending Tina's random number generators.
// We implement a exclusive or lagged Fibonacci generator. 
//
//   r_i=r_(i-q) xor r_(i-p); p>q
//
// Sequence splitting and leapfrog method are implemented in a very stupid 
// way. Unused values are just thrown away. Don't use this generator in
// your applications.

// the template parametr determine the lags
template<long p1, long q1>
class Fibonacci : public TRNG::RNG<Fibonacci<p1, q1> > {
private:
  std::vector<long> r;
  long p, q, pointer, steps;
  TRNG::ParkMiller R_init;

  void backward(void) {
    for (long i=0l; i<steps; ++i) {
      r[pointer]=r[pointer]^r[(pointer+q)%p];
      pointer=(pointer-1l+p)%p;
    }
  }

public:
  // TRNG::user1_t for a user implemeneted rng
  static const TRNG::RNG_type type=TRNG::user1_t;

  const char * name(void) {
    return "Fibonacci";
  }

  void reset(void) {
    steps=1l;
    max_val=0x7fffffffl;
    max_val2=max_val/2l;
  }

  // this is the initialization procedure described in Fushimi and Tezuka,
  // Comm. ACM, 26 (1983) 516 which ensures that the resulting sequence is
  // P/32-distributed
  void seed(long s) {
    std::vector<unsigned long int> e(p1);
    std::vector<unsigned short int> a(2*p1);
    // construct P linearly independent basis vectors
    for (int i=0; i<p1; i++) {
      int k=31*i/p1;
      e[i]=static_cast<unsigned long int>(s) << k; // zeros to right of bit k
      e[i]=e[i] | ( 1l << k ); // and a one at bit k
    }
    // construct 2P-1 coefficient bits
    for (int i=0; i<p1; i++) 
      a[i]=R_init.boolean() ? 1 : 0;
    for (int i=p1; i<2*p1; i++)
      a[i]=a[i-p1]^a[i-q1];
    // construct first P-1 entries (matrix seed) by
    // combining basis vectors according to coefficient bits
    for (int i=0; i<p1; i++) {
      r[i]=0;
      for (int j=0; j<p1; j++) 
	if (a[i+j]) 
	  r[i]=r[i]^e[j];
    }
    pointer=p-1l;
  }
  
  // calculate the next random number
  long rand(void) {
    long t;
    ++pointer;
    if (pointer==p)
      pointer=0l;
    r[pointer]=r[pointer]^r[(pointer+q)<p ? (pointer+q) : (pointer+q-p)];
    t=r[pointer];
    for (long i=1l; i<steps; ++i) {
      ++pointer;
      if (pointer==p)
	pointer=0l;
      r[pointer]=r[pointer]^r[(pointer+q)<p ? (pointer+q) : (pointer+q-p)];
    }
    return t;
  }
  
  void split(long s, long n) {
    if (s<1l || n>s || n<0l)
      throw TRNG::error("invalid arguments for Fibonacci::split");
    if (s>1l) {
      for (long i=0l; i<n; ++i)
	rand();
      steps*=s;
    }
  }

  void jump2(long s) {
    if (s<0l || s>63l)
      throw TRNG::error("invalid argument for Fibonacci::split");
    unsigned long long to=1ull<<s;
    for (unsigned long long i=0ull; i<to; ++i)
      rand();
  }

  void save_status(std::vector<long> &s) {
    s.resize(p+5);
    s[0]=type;
    s[1]=q;
    s[2]=p;
    s[3]=pointer;
    s[4]=steps;
    for (int i=0; i<p; ++i)
      s[i+5]=r[i];
  }

  void load_status(const std::vector<long> &s) {
    if (s[0]!=type)
      throw TRNG::error("Fibonacci::load_status wrong parameter");
    q=s[1];
    p=s[2];
    pointer=s[3];
    steps=s[4];
    r.resize(p);
      for (int i=0; i<p; ++i)
	r[i]=s[i+5];
  }

  Fibonacci & Fibonacci::operator=(TRNG::RNG<Fibonacci> &other) {
    if (this!=&other) {
      std::vector<long> s;
      other.save_status(s);
      load_status(s);
    }
    return *this;
  }
  
  Fibonacci(long seed_=0l) : r() {
    if (p1<0l || q1<0l || p1==q1)
      throw TRNG::error("bad arguments for Fibonacci::Fibonacci");
    if (p1>q1) {
      p=p1;
      q=q1;
    } else {
      p=q1;
      q=p1;
    }
    r.resize(p);
    reset();
    seed(seed_);
  }

  virtual ~Fibonacci() {};
};


#endif
